# built-in
from typing import Optional

# app
from .common import TOKENS, Extractor, Token
from .value import UNKNOWN, get_value


get_asserts = Extractor()


@get_asserts.register(*TOKENS.ASSERT)
def handle_assert(expr) -> Optional[Token]:
    value = get_value(expr=expr.test)
    if value is UNKNOWN:
        return None
    if value:
        return None
    return Token(value=value, line=expr.lineno, col=expr.col_offset + 7)



# built-in
import sys
from argparse import ArgumentParser
from contextlib import contextmanager
from importlib import import_module
from pathlib import Path
from textwrap import indent
from traceback import format_exception
from typing import Iterator, Sequence, TextIO

# app
from .._testing import cases
from ..linter._contract import Category
from ..linter._extractors.pre import format_call_args
from ..linter._func import Func


COLORS = dict(
    red='\033[91m',
    green='\033[92m',
    yellow='\033[93m',
    blue='\033[94m',
    magenta='\033[95m',
    end='\033[0m',
)


@contextmanager
def sys_path(path: Path):
    path = str(path)
    sys.path.insert(0, path)
    try:
        yield
    finally:
        if sys.path[0] == path:
            del sys.path[0]


def has_pure_contract(func: Func) -> bool:
    for contract in func.contracts:
        if contract.category == Category.PURE:
            return True
        if contract.category == Category.HAS and not contract.args:
            return True
    return False


def get_func_names(path: Path) -> Iterator[str]:
    for func in Func.from_path(path=path):
        if has_pure_contract(func=func):
            yield func.name


def print_exception(stream: TextIO) -> None:
    lines = format_exception(*sys.exc_info())
    text = indent(text=''.join(lines), prefix='    ')
    text = '{red}{text}{end}'.format(text=text, **COLORS)
    print(text, file=stream)


def run_tests(path: Path, root: Path, count: int, stream: TextIO = sys.stdout) -> int:
    names = list(get_func_names(path=path))
    if not names:
        return 0
    print('{magenta}running {path}{end}'.format(path=path, **COLORS), file=stream)
    module_name = '.'.join(path.relative_to(root).with_suffix('').parts)
    with sys_path(path=root):
        module = import_module(module_name)
    failed = 0
    for func_name in names:
        print('  {blue}running {name}{end}'.format(name=func_name, **COLORS), file=stream)
        func = getattr(module, func_name)
        for case in cases(func=func, count=count):
            try:
                case()
            except Exception:
                line = '    {yellow}{name}({args}){end}'.format(
                    name=func_name,
                    args=format_call_args(args=case.args, kwargs=case.kwargs),
                    **COLORS,
                )
                print(line, file=stream)
                print_exception(stream=stream)
                failed += 1
                break
    return failed


def test_command(
    argv: Sequence[str], root: Path = None, stream: TextIO = sys.stdout,
) -> int:
    if root is None:  # pragma: no cover
        root = Path()
    parser = ArgumentParser(prog='python3 -m deal test')
    parser.add_argument('--count', type=int, default=50)
    parser.add_argument('paths', nargs='+')
    args = parser.parse_args(argv)

    failed = 0
    for path in args.paths:
        failed += run_tests(
            path=Path(path),
            root=root,
            count=args.count,
            stream=stream,
        )
    return failed



# built-in
import ast
from typing import Iterator, Sequence

# external
import astroid

# app
from .common import Extractor, Token, infer
from .contracts import get_contracts
from .value import UNKNOWN, get_value


get_pre = Extractor()


@get_pre.register(astroid.Call)
def handle_call(expr: astroid.Call) -> Iterator[Token]:
    # app
    from .._contract import Category, Contract

    args = []
    for subnode in expr.args:
        value = get_value(expr=subnode)
        if value is UNKNOWN:
            return
        args.append(value)

    kwargs = {}
    for subnode in (expr.keywords or ()):
        value = get_value(expr=subnode.value)
        if value is UNKNOWN:
            return
        kwargs[subnode.arg] = value

    for func in infer(expr.func):
        if type(func) is not astroid.FunctionDef:
            continue
        if not func.decorators:
            continue
        code = 'def f({}):0'.format(func.args.as_string())
        func_args = ast.parse(code).body[0].args  # type: ignore

        for category, contract_args in get_contracts(func.decorators.nodes):
            if category != 'pre':
                continue

            contract = Contract(
                args=contract_args,
                category=Category.PRE,
                func_args=func_args,
            )
            try:
                result = contract.run(*args, **kwargs)
            except NameError:
                continue
            if result is False or type(result) is str:
                msg = format_call_args(args, kwargs)
                yield Token(value=msg, line=expr.lineno, col=expr.col_offset)


def format_call_args(args: Sequence, kwargs: dict) -> str:
    sep = ', '
    args_s = sep.join(map(repr, args))
    kwargs_s = sep.join(['{}={!r}'.format(k, v) for k, v in kwargs.items()])
    if args and kwargs:
        return args_s + sep + kwargs_s
    return args_s + kwargs_s





class _State:
    __slots__ = ('debug', )
    debug: bool

    def __init__(self):
        self.reset()

    def reset(self) -> None:
        """Restore contracts switch to default.

        All contracts are disabled on production by default.
        See [runtime](../basic/runtime.md) documentation.
        """
        self.debug = __debug__

    def enable(self) -> None:
        """Enable all contracts.
        """
        self.debug = True

    def disable(self) -> None:
        """Disable all contracts.
        """
        self.debug = False


state = _State()
reset = state.reset
enable = state.enable
disable = state.disable



# This file is excluded from coverage.

# project
from deal import ContractError
from deal._decorators.base import Base


# will be filled from the linter
contract = ...
func = ...

base = Base(validator=contract)  # type: ignore
if func is not Ellipsis:
    base.function = func

try:
    base.validate(*args, **kwargs)  # type: ignore  # noqa: F821
except ContractError as exc:
    result = False
    if exc.args:
        result = exc.args[0]
else:
    result = True



# built-in
from argparse import ArgumentParser
from pathlib import Path
from typing import Sequence

# app
from ..linter import StubsManager, generate_stub


def stub_command(argv: Sequence[str]) -> int:
    parser = ArgumentParser(prog='python3 -m deal stub')
    parser.add_argument('--iterations', type=int, default=1)
    parser.add_argument('paths', nargs='+')
    args = parser.parse_args(argv)

    paths = [Path(path) for path in args.paths]
    roots = list(StubsManager.default_paths) + list(set(paths))
    stubs = StubsManager(paths=roots)

    for _ in range(args.iterations):
        for path in paths:
            generate_stub(path=path, stubs=stubs)
    return 0



# built-in
import ast
import builtins
import enum
from pathlib import Path

# external
import astroid


TEMPLATE = (Path(__file__).parent / '_template.py').read_text()


class Category(enum.Enum):
    HAS = 'has'
    PRE = 'pre'
    POST = 'post'
    RAISES = 'raises'
    PURE = 'pure'


class Contract:
    __slots__ = ('args', 'category', 'func_args')

    def __init__(self, args, category: Category, func_args: ast.arguments):
        self.args = args
        self.category = category
        self.func_args = func_args

    @property
    def body(self) -> ast.AST:
        contract = self.args[0]
        # convert astroid node to ast node
        if hasattr(contract, 'as_string'):
            contract = self._resolve_name(contract)
            contract = ast.parse(contract.as_string()).body[0]
        return contract

    @staticmethod
    def _resolve_name(contract):
        if not isinstance(contract, astroid.Name):
            return contract
        definitions = contract.lookup(contract.name)[1]
        if not definitions:
            return contract
        definition = definitions[0]
        if isinstance(definition, astroid.FunctionDef):
            return definition
        if isinstance(definition, astroid.AssignName):
            return definition.parent.value
        # resolved into something tricky, live with it
        return contract  # pragma: no cover

    @property
    def exceptions(self) -> list:
        # app
        from ._extractors import get_name

        excs = []
        for expr in self.args:
            name = get_name(expr)
            if not name:
                continue
            exc = getattr(builtins, name, name)
            excs.append(exc)
        return excs

    @property
    def bytecode(self):
        module = ast.parse(TEMPLATE)

        # inject function signature
        func = ast.parse('lambda:0').body[0].value
        func.args = self.func_args
        module.body[3].value = func

        # inject contract
        contract = self.body
        if isinstance(contract, ast.FunctionDef):
            # if contract is function, add it's definition and assign it's name
            # to `contract` variable.
            module.body = [contract] + module.body
            module.body[3].value = ast.Name(
                id=contract.name,
                lineno=1,
                col_offset=1,
                ctx=ast.Load(),
            )
        else:
            if isinstance(contract, ast.Expr):
                contract = contract.value
            module.body[2].value = contract

        return compile(module, filename='<ast>', mode='exec')

    def run(self, *args, **kwargs):
        globals = dict(args=args, kwargs=kwargs)
        exec(self.bytecode, globals)
        return globals['result']

    def __repr__(self) -> str:
        return '{name}({category})'.format(
            name=type(self).__name__,
            category=self.category.value,
        )



# built-in
import inspect
from asyncio import iscoroutinefunction
from contextlib import suppress
from functools import update_wrapper
from typing import Callable, Generic, TypeVar

# external
import vaa

# app
from .._exceptions import ContractError
from .._state import state
from .._types import ExceptionType


#: We use this type in many other subclasses of `Base` decorator.
_CallableType = TypeVar('_CallableType', bound=Callable)


class Base(Generic[_CallableType]):
    exception: ExceptionType = ContractError
    function: _CallableType  # pytype: disable=not-supported-yet

    def __init__(self, validator, *, message: str = None,
                 exception: ExceptionType = None):
        """
        Step 1. Set contract (validator).
        """
        self.validator = self._make_validator(validator, message=message)
        if exception:
            self.exception = exception
        if message:
            self.exception = self.exception(message)    # type: ignore

    @staticmethod
    def _make_validator(validator, message: str = None):
        if validator is None:
            return None
        # implicitly wrap in vaa all external validators
        with suppress(TypeError):
            return vaa.wrap(validator, simple=False)

        # implicitly wrap in vaa.simple only funcs with one `_` argument.
        if inspect.isfunction(validator):
            params = inspect.signature(validator).parameters
            if set(params) == {'_'}:
                return vaa.simple(validator, error=message)

        return validator

    def validate(self, *args, **kwargs) -> None:
        """
        Step 4. Process contract (validator)
        """

        if hasattr(self.validator, 'is_valid'):
            self._vaa_validation(*args, **kwargs)
        else:
            self._simple_validation(*args, **kwargs)

    def _vaa_validation(self, *args, **kwargs) -> None:
        params = kwargs.copy()

        # if it is a decorator for a function, convert positional args into named ones.
        if hasattr(self, 'function'):
            # detect original function
            function = self.function
            while hasattr(function, '__wrapped__'):
                function = function.__wrapped__     # type: ignore
            # assign *args to real names
            kwargs.pop('result', None)
            params.update(inspect.getcallargs(function, *args, **kwargs))
            # drop args-kwargs, we already put them on the right places
            for bad_name in ('args', 'kwargs'):
                if bad_name in params and bad_name not in kwargs:
                    del params[bad_name]

        # validate
        validator = self.validator(data=params)
        if validator.is_valid():
            return

        # if no errors returned, raise the default exception
        errors = validator.errors
        if not errors:
            raise self.exception

        # Flatten single error without field to one simple str message.
        # This is for better readability of simple validators.
        if type(errors) is list:  # pragma: no cover
            if type(errors[0]) is vaa.Error:
                if len(errors) == 1:
                    if errors[0].field is None:
                        errors = errors[0].message

        # raise errors
        if isinstance(self.exception, Exception):
            raise type(self.exception)(errors)
        raise self.exception(errors)

    def _simple_validation(self, *args, **kwargs) -> None:
        validation_result = self.validator(*args, **kwargs)
        # is invalid (validator returns error message)
        if isinstance(validation_result, str):
            if isinstance(self.exception, Exception):
                raise type(self.exception)(validation_result)
            raise self.exception(validation_result)
        # is valid (truely result)
        if validation_result:
            return
        # is invalid (falsy result)
        raise self.exception

    @property
    def enabled(self) -> bool:
        return state.debug

    def __call__(self, function: _CallableType) -> _CallableType:
        """
        Step 2. Return wrapped function.
        """
        self.function = function

        def wrapped(*args, **kwargs):
            if self.enabled:
                return self.patched_function(*args, **kwargs)
            else:
                return function(*args, **kwargs)

        async def async_wrapped(*args, **kwargs):
            if self.enabled:
                return await self.async_patched_function(*args, **kwargs)
            else:
                return await function(*args, **kwargs)

        def wrapped_generator(*args, **kwargs):
            if self.enabled:
                yield from self.patched_generator(*args, **kwargs)
            else:
                yield from function(*args, **kwargs)

        if iscoroutinefunction(function):
            new_callable = update_wrapper(async_wrapped, function)
        elif inspect.isgeneratorfunction(function):
            new_callable = update_wrapper(wrapped_generator, function)
        else:
            new_callable = update_wrapper(wrapped, function)
        return new_callable  # type: ignore

    def patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError

    async def async_patched_function(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError

    def patched_generator(self, *args, **kwargs):
        """
        Step 3. Wrapped function calling.
        """
        raise NotImplementedError





class ContractError(AssertionError):
    pass


class PreContractError(ContractError):
    pass


class PostContractError(ContractError):
    pass


class InvContractError(ContractError):
    pass


class RaisesContractError(ContractError):
    pass


class ReasonContractError(ContractError):
    pass


class MarkerError(ContractError):
    pass


class OfflineContractError(MarkerError):
    pass


class SilentContractError(MarkerError):
    pass
